"""Contains basic recorders."""
from typing import Any
from brian2 import PopulationRateMonitor, StateMonitor, SpikeMonitor, Quantity
import numpy as np
from nptyping import NDArray

from cleosim.base import Recorder


class RateRecorder(Recorder):
    """Records firing rate from a single neuron.

    Firing rate comes from Brian's :class:`~brian2.monitors.ratemonitor.PopulationRateMonitor`"""

    def __init__(self, name: str, index: int):
        """
        Parameters
        ----------
        name : str
            Unique device name
        index : int
            index of neuron to record
        """
        super().__init__(name)
        self.i = index
        self.mon = None

    def connect_to_neuron_group(self, neuron_group):
        self.mon = PopulationRateMonitor(neuron_group[self.i])
        self.brian_objects.add(self.mon)

    def get_state(self):
        try:
            return self.mon.rate[-1]
        except IndexError:
            return 0


class VoltageRecorder(Recorder):
    """Records the voltage of a single neuron group."""

    def __init__(self, name: str, voltage_var_name: str = "v"):
        """
        Parameters
        ----------
        name : str
            Unique device name
        voltage_var_name : str, optional
            Name of variable representing membrane voltage, by default "v"
        """
        super().__init__(name)
        self.voltage_var_name = voltage_var_name
        self.mon = None

    def connect_to_neuron_group(self, neuron_group):
        if self.mon is not None:
            raise UserWarning(
                "Recorder was already connected to a neuron group. "
                "Can only record from one at a time."
            )
        self.mon = StateMonitor(neuron_group, self.voltage_var_name, record=True)
        self.brian_objects.add(self.mon)

    def get_state(self) -> Quantity:
        """
        Returns
        -------
        Quantity
            Current voltage of target neuron group
        """
        try:
            return self.mon.v[:, -1]
        except IndexError:
            return None


class GroundTruthSpikeRecorder(Recorder):
    """Reports the number of spikes seen since last queried for each neuron.

    This amounts effectively to the number of spikes per control period.
    Note: this will only work for one neuron group at the moment.
    """

    def __init__(self, name):
        super().__init__(name)
        self.mon = None
        self.num_spikes_seen = 0

    def connect_to_neuron_group(self, neuron_group):
        if self.mon is not None:
            raise UserWarning(
                "Recorder was already connected to a neuron group. "
                "Can only record from one at a time."
            )
        self.mon = SpikeMonitor(neuron_group)
        self.brian_objects.add(self.mon)
        self.out_template = np.zeros(len(neuron_group))

    def get_state(self) -> NDArray[(Any,), np.uint]:
        """
        Returns
        -------
        NDArray[(n_neurons,), np.uint]
            n_neurons-length array with spike counts over the latest
            control period.
        """
        num_new_spikes = len(self.mon.t) - self.num_spikes_seen
        self.num_spikes_seen += num_new_spikes
        if len(self.out_template) == 1:
            out = np.array([num_new_spikes])
        else:
            out = self.out_template.copy()
            for spike_i in self.mon.i[-num_new_spikes:]:
                out[spike_i] += 1
        return out



"""Tools for visualizing models and simulations"""
from __future__ import annotations
from typing import Tuple, Any, Union
from collections.abc import Iterable
from matplotlib.artist import Artist

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.animation as anim
from mpl_toolkits.mplot3d import Axes3D
from brian2 import mm, Unit, NeuronGroup, ms, NetworkOperation, Quantity, SpikeMonitor

from cleosim.base import CLSimulator, InterfaceDevice

_neuron_alpha = 0.2


class VideoVisualizer(InterfaceDevice):
    """Device for visualizing a simulation.

    Must be injected after all other devices and before the simulation
    is run."""

    def __init__(
        self,
        devices: Iterable[Union[InterfaceDevice, Tuple[InterfaceDevice, dict]]] = "all",
        dt: Quantity = 1 * ms,
    ) -> None:
        """
        Parameters
        ----------
        devices : Iterable[Union[InterfaceDevice, Tuple[InterfaceDevice, dict]]], optional
            list of devices or (device, vis_kwargs) tuples to include in the plot,
            just as in the :func:`plot` function, by default "all", which will
            include all recorders and stimulators currently injected when this
            visualizer is injected into the simulator.
        dt : Brian 2 temporal Quantity, optional
            length of each frame---that is, every `dt` the visualizer takes a
            snapshot of the network, by default 1*ms
        """
        super().__init__("video_visualizer")
        self.neuron_groups = []
        self._spike_mons = []
        self._num_old_spikes = []
        self.devices = devices
        self.dt = dt
        # store data to generate video
        self._value_per_device_per_frame = []
        self._i_spikes_per_ng_per_frame: list[list[np.ndarray]] = []

    def init_for_simulator(self, simulator: CLSimulator):
        if self.devices == "all":
            self.devices = list(self.sim.recorders.values())
            self.devices.extend(list(self.sim.stimulators.values()))
        # network op
        def snapshot(t):
            i_spikes_per_ng = [
                self._new_spikes_for_ng(i_ng) for i_ng in range(len(self.neuron_groups))
            ]
            self._i_spikes_per_ng_per_frame.append(i_spikes_per_ng)
            device_values = []
            for device in self.devices:
                try:
                    device_values.append(device.value)
                # not all devices (recorders!) have a value or any changing state to plot
                except AttributeError:
                    device_values.append(None)
            self._value_per_device_per_frame.append(device_values)

        simulator.network.add(NetworkOperation(snapshot, dt=self.dt))

    def connect_to_neuron_group(self, neuron_group: NeuronGroup, **kwparams) -> None:
        self.neuron_groups.append(neuron_group)
        mon = SpikeMonitor(neuron_group)
        self._spike_mons.append(mon)
        self.brian_objects.add(mon)
        self._num_old_spikes.append(0)

    def generate_Animation(
        self, plotargs: dict, slowdown_factor: float = 10, **figargs: Any
    ) -> anim.Animation:
        """Create a matplotlib Animation object from the recorded simulation

        Parameters
        ----------
        plotargs : dict
            dictionary of arguments as taken by :func:`plot`. can include
            `xlim`, `ylim`, `zlim`, `colors`, `axis_scale_unit`, `invert_z`,
            and/or `scatterargs`. neuron groups and devices are
            automatically added and **figargs are specified separately.
        slowdown_factor : float, optional
            how much slower the animation will be rendered, as a multiple of
            real-time, by default 10
        **figargs : Any, optional
            keyword arguments passed to plt.figure(), such as figsize

        Returns
        -------
        :class:`matplotlib.animation.Animation`
            An Animation object capturing the desired visualization.
            See matplotlib's docs for saving and rendering options.
        """
        interval_ms = self.dt / ms * slowdown_factor
        self.fig = plt.figure(**figargs)
        self.ax = self.fig.add_subplot(111, projection="3d")
        neuron_artists, device_artists = _plot(
            self.ax,
            self.neuron_groups,
            devices=self.devices,
            **plotargs,
        )

        def update(i):
            device_values = self._value_per_device_per_frame[i]
            updated_artists = []
            for device, artists, value in zip(
                self.devices, device_artists, device_values
            ):
                updated_artists_for_device = device.update_artists(artists, value)
                updated_artists.extend(updated_artists_for_device)
            self._update_neuron_artists_for_frame(neuron_artists, i)
            return updated_artists + neuron_artists

        return anim.FuncAnimation(
            self.fig,
            update,
            range(len(self._value_per_device_per_frame)),
            interval=interval_ms,
            blit=True,
        )

    def _update_neuron_artists_for_frame(self, artists, i_frame):
        i_spikes_per_ng = self._i_spikes_per_ng_per_frame[i_frame]
        # loop over neuron groups/artists
        for i_spikes, ng, artist in zip(i_spikes_per_ng, self.neuron_groups, artists):
            spike_counts = self._spikes_i_to_count_for_ng(i_spikes, ng)
            artist.set_alpha(None)  # remove alpha defined at collection level
            rgba = artist.get_edgecolor()
            alpha = np.zeros(ng.N)
            alpha[spike_counts == 0] = _neuron_alpha
            alpha[spike_counts > 0] = 1
            rgba[:, 3] = alpha
            # warning: this doesn't work. gets order wrong: artist.set_alpha(alpha)
            artist.set_color(rgba)

    def _spikes_i_to_count_for_ng(self, i_spikes, ng):
        counts = np.zeros(ng.N)
        counts[i_spikes] = 1
        return counts

    def _new_spikes_for_ng(self, i_ng):
        mon = self._spike_mons[i_ng]
        num_old = self._num_old_spikes[i_ng]
        new_i_spikes = mon.i[num_old:]
        self._num_old_spikes[i_ng] = len(mon.i)
        return new_i_spikes


def _plot(
    ax: Axes3D,
    neuron_groups: NeuronGroup,
    xlim: Tuple[float, float] = None,
    ylim: Tuple[float, float] = None,
    zlim: Tuple[float, float] = None,
    colors: Iterable = None,
    axis_scale_unit: Unit = mm,
    devices: Iterable[Union[InterfaceDevice, Tuple[InterfaceDevice, dict]]] = [],
    invert_z: bool = True,
    scatterargs: dict = {},
) -> tuple[list[Artist], list[Artist]]:
    for ng in neuron_groups:
        for dim in ["x", "y", "z"]:
            if not hasattr(ng, dim):
                raise ValueError(f"{ng.name} does not have dimension {dim} defined.")

    assert colors is None or len(colors) == len(neuron_groups)
    neuron_artists = []
    for i in range(len(neuron_groups)):
        ng = neuron_groups[i]
        xyz = [ng.x / axis_scale_unit, ng.y / axis_scale_unit, ng.z / axis_scale_unit]
        # mask neurons outside desired lims:
        for i_dim, lim in enumerate([xlim, ylim, zlim]):
            if lim is not None:
                xyz[i_dim] = np.ma.masked_array(
                    xyz[i_dim],
                    np.logical_or(xyz[i_dim] < lim[0], xyz[i_dim] > lim[1]),
                    dtype=float,
                )
        kwargs = {"label": ng.name, "alpha": _neuron_alpha}
        if colors is not None:
            kwargs["color"] = colors[i]
        kwargs.update(scatterargs)
        neuron_artists.append(ax.scatter(*xyz, **kwargs))
        ax.set_xlabel(f"x ({axis_scale_unit._dispname})")
        ax.set_ylabel(f"y ({axis_scale_unit._dispname})")
        ax.set_zlabel(f"z ({axis_scale_unit._dispname})")

    xlim = ax.get_xlim() if xlim is None else xlim
    ylim = ax.get_ylim() if ylim is None else ylim
    zlim = ax.get_zlim() if zlim is None else zlim

    ax.set(xlim=xlim, ylim=ylim, zlim=zlim)
    z_aspect = zlim[0] - zlim[1] if invert_z else zlim[1] - zlim[0]
    ax.set_box_aspect((xlim[1] - xlim[0], ylim[1] - ylim[0], z_aspect))

    ax.legend()

    device_artists = []
    for item in devices:
        if isinstance(item, InterfaceDevice):
            device, plotargs = item, {}
        elif type(item) == tuple:
            device, plotargs = item
        else:
            raise Exception(
                "InterfaceDevice or (InterfaceDevice, dict) expected."
                f"Instead received {item}."
            )
        device_artists.append(device.add_self_to_plot(ax, axis_scale_unit, **plotargs))

    return neuron_artists, device_artists


def plot(
    *neuron_groups: NeuronGroup,
    xlim: Tuple[float, float] = None,
    ylim: Tuple[float, float] = None,
    zlim: Tuple[float, float] = None,
    colors: Iterable = None,
    axis_scale_unit: Unit = mm,
    devices: Iterable[Union[InterfaceDevice, Tuple[InterfaceDevice, dict]]] = [],
    invert_z: bool = True,
    scatterargs: dict = {},
    **figargs: Any,
) -> None:
    """Visualize neurons and interface devices

    Parameters
    ----------
    xlim : Tuple[float, float], optional
        xlim for plot, determined automatically by default
    ylim : Tuple[float, float], optional
        ylim for plot, determined automatically by default
    zlim : Tuple[float, float], optional
        zlim for plot, determined automatically by default
    colors : Iterable, optional
        colors, one for each neuron group, automatically determined by default
    axis_scale_unit : Unit, optional
        Brian unit to scale lim params, by default mm
    devices : Iterable[Union[InterfaceDevice, Tuple[InterfaceDevice, dict]]], optional
        devices to add to the plot or (device, kwargs) tuples.
        :meth:`~cleosim.InterfaceDevice.add_self_to_plot` is called
        for each, using the kwargs dict if given. By default []
    invert_z : bool, optional
        whether to invert z-axis, by default True to reflect the convention
        that +z represents depth from cortex surface
    scatterargs : dict, optional
        arguments passed to plt.scatter() for each neuron group, such as marker
    **figargs : Any, optional
        keyword arguments passed to plt.figure(), such as figsize

    Raises
    ------
    ValueError
        When neuron group doesn't have x, y, and z already defined
    """
    fig = plt.figure(**figargs)
    ax = fig.add_subplot(111, projection="3d")
    _plot(
        ax,
        neuron_groups,
        xlim,
        ylim,
        zlim,
        colors,
        axis_scale_unit,
        devices,
        invert_z,
        scatterargs,
    )
    return fig, ax



"""Contains definitions for essential, base classes."""

from __future__ import annotations
from abc import ABC, abstractmethod
from typing import Any

from brian2 import (
    NeuronGroup,
    Subgroup,
    Network,
    NetworkOperation,
    defaultclock,
    ms,
    Unit,
    Quantity,
)
from matplotlib import pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
from matplotlib.artist import Artist


class InterfaceDevice(ABC):
    """Base class for devices to be injected into the network"""

    name: str
    """Unique identifier for device.
    Used as a key in output/input dicts
    """
    brian_objects: set
    """All the Brian objects added to the network by this device.
    Must be kept up-to-date in :meth:`connect_to_neuron_group` and
    other functions so that those objects can be automatically added
    to the network when the device is injected.
    """
    sim: CLSimulator
    """The simulator the device is injected into
    """

    def __init__(self, name: str) -> None:
        """
        Parameters
        ----------
        name : str
            Unique identifier for the device.
        """
        self.name = name
        self.brian_objects = set()
        self.sim = None

    def init_for_simulator(self, simulator: CLSimulator) -> None:
        """Initialize device for simulator on initial injection

        This function is called only the first time a device is
        injected into a simulator and performs any operations
        that are independent of the individual neuron groups it
        is connected to.

        Parameters
        ----------
        simulator : CLSimulator
            simulator being injected into
        """
        pass

    @abstractmethod
    def connect_to_neuron_group(self, neuron_group: NeuronGroup, **kwparams) -> None:
        """Connect device to given `neuron_group`.

        If your device introduces any objects which Brian must
        keep track of, such as a NeuronGroup, Synapses, or Monitor,
        make sure to add these to `self.brian_objects`.

        Parameters
        ----------
        neuron_group : NeuronGroup
        **kwparams : optional, passed from `inject_recorder` or
            `inject_stimulator`
        """
        pass

    def add_self_to_plot(self, ax: Axes3D, axis_scale_unit: Unit, **kwargs) -> list[Artist]:
        """Add device to an existing plot

        Should only be called by :func:`~cleosim.visualization.plot`.

        Parameters
        ----------
        ax : Axes3D
            The existing matplotlib Axes object
        axis_scale_unit : Unit
            The unit used to label axes and define chart limits
        **kwargs : optional

        Returns
        -------
        list[Artist]
            A list of artists used to render the device. Needed for use
            in conjunction with :class:`~cleosim.visualization.VideoVisualizer`.
        """
        return []

    def update_artists(artists: list[Artist], *args, **kwargs) -> list[Artist]:
        """Update the artists used to render the device

        Used to set the artists' state at every frame of a video visualization.
        The current state would be passed in `*args` or `**kwargs`

        Parameters
        ----------
        artists : list[Artist]
            the artists used to render the device originally, i.e.,
            which were returned from the first :meth:`add_self_to_plot` call.

        Returns
        -------
        list[Artist]
            The artists that were actually updated. Needed for efficient
            blit rendering, where only updated artists are re-rendered.
        """
        return []


class IOProcessor(ABC):
    """Abstract class for implementing sampling, signal processing and control

    This must be implemented by the user with their desired closed-loop
    use case, though most users will find the :func:`~processing.LatencyIOProcessor`
    class more useful, since delay handling is already defined.
    """

    sample_period_ms: float
    """Determines how frequently the processor takes samples"""

    @abstractmethod
    def is_sampling_now(self, time) -> bool:
        """Determines whether the processor will take a sample at this timestep.

        Parameters
        ----------
        time : Brian 2 temporal Unit
            Current timestep.

        Returns
        -------
        bool
        """
        pass

    @abstractmethod
    def put_state(self, state_dict: dict, time) -> None:
        """Deliver network state to the :class:`IOProcessor`.

        Parameters
        ----------
        state_dict : dict
            A dictionary of recorder measurements, as returned by
            :func:`~cleosim.CLSimulator.get_state()`
        time : brian2 temporal Unit
            The current simulation timestep. Essential for simulating
            control latency and for time-varying control.
        """
        pass

    @abstractmethod
    def get_ctrl_signal(self, time) -> dict:
        """Get per-stimulator control signal from the :class:`~cleosim.IOProcessor`.

        Parameters
        ----------
        time : Brian 2 temporal Unit
            Current timestep

        Returns
        -------
        dict
            A {'stimulator_name': value} dictionary for updating stimulators.
        """
        pass

    def reset(self, **kwargs) -> None:
        pass


class Recorder(InterfaceDevice):
    """Device for taking measurements of the network."""

    @abstractmethod
    def get_state(self) -> Any:
        """Return current measurement."""
        pass

    def reset(self, **kwargs) -> None:
        """Reset the recording device to a neutral state"""
        pass


class Stimulator(InterfaceDevice):
    """Device for manipulating the network"""

    value: Any
    """The current value of the stimulator device"""
    default_value: Any
    """The default value of the device---used on initialization and on :meth:`~reset`"""
    t_ms: list[float]
    """Times stimulator was updated, stored if :attr:`save_history`"""
    values: list[Any]
    """Values taken by the stimulator at each :meth:`~update` call, 
    stored if :attr:`save_history`"""
    save_history: bool
    """Determines whether :attr:`t_ms` and :attr:`values` are recorded"""

    def __init__(self, name: str, default_value, save_history: bool = False) -> None:
        """
        Parameters
        ----------
        name : str
            Unique device name used in :meth:`CLSimulator.update_stimulators`
        default_value : any
            The stimulator's default value
        """
        super().__init__(name)
        self.value = default_value
        self.default_value = default_value
        self.save_history = save_history

    def init_for_simulator(self, simulator: CLSimulator) -> None:
        super().init_for_simulator(simulator)
        self._init_saved_vars()

    def _init_saved_vars(self):
        if self.save_history:
            self.t_ms = [self.sim.network.t / ms]
            self.values = [self.default_value]

    def update(self, ctrl_signal) -> None:
        """Set the stimulator value.

        By default this simply sets `value` to `ctrl_signal`.
        You will want to implement this method if your
        stimulator requires additional logic. Use super.update(self, value)
        to preserve the self.value attribute logic

        Parameters
        ----------
        ctrl_signal : any
            The value the stimulator is to take.
        """
        self.value = ctrl_signal
        if self.save_history:
            self.t_ms.append(self.sim.network.t / ms)
            self.values.append(self.value)

    def reset(self, **kwargs) -> None:
        """Reset the stimulator device to a neutral state"""
        self._init_saved_vars()


class CLSimulator:
    """The centerpiece of cleosim. Integrates simulation components and runs."""

    io_processor: IOProcessor
    network: Network
    recorders = "set[Recorder]"
    stimulators = "set[Stimulator]"
    _processing_net_op: NetworkOperation
    _net_store_name: str = "cleosim default"

    def __init__(self, brian_network: Network) -> None:
        """
        Parameters
        ----------
        brian_network : Network
            The Brian network forming the core model
        """
        self.network = brian_network
        self.stimulators = {}
        self.recorders = {}
        self.io_processor = None
        self._processing_net_op = None

    def inject_device(
        self, device: InterfaceDevice, *neuron_groups: NeuronGroup, **kwparams: Any
    ) -> None:
        """Inject InterfaceDevice into the network, connecting to specified neurons.

        Calls :meth:`~InterfaceDevice.connect_to_neuron_group` for each group with
        kwparams and adds the device's :attr:`~InterfaceDevice.brian_objects`
        to the simulator's :attr:`network`.

        Automatically called by :meth:`inject_recorder` and :meth:`inject_stimulator`.

        Parameters
        ----------
        device : InterfaceDevice
            Device to inject
        """
        if len(neuron_groups) == 0:
            raise Exception("Injecting stimulator for no neuron groups is meaningless.")
        for ng in neuron_groups:
            if type(ng) == NeuronGroup:
                if ng not in self.network.objects:
                    raise Exception(
                        f"Attempted to connect device {device.name} to neuron group "
                        f"{ng.name}, which is not part of the simulator's network."
                    )
            elif type(ng) == Subgroup:
                # must look at sorted_objects because ng.source is unhashable
                if ng.source not in self.network.sorted_objects:
                    raise Exception(
                        f"Attempted to connect device {device.name} to neuron group "
                        f"{ng.source.name}, which is not part of the simulator's network."
                    )
            if device.sim not in [None, self]:
                raise Exception(
                    f"Attempted to inject device {device.name} into {self}, "
                    f"but it was previously injected into {device.sim}. "
                    "Each device can only be injected into one CLSimulator."
                )
            if device.sim is None:
                device.sim = self
                device.init_for_simulator(self)
            device.connect_to_neuron_group(ng, **kwparams)
        for brian_object in device.brian_objects:
            self.network.add(brian_object)
        self.network.store(self._net_store_name)

    def inject_stimulator(
        self, stimulator: Stimulator, *neuron_groups: NeuronGroup, **kwparams
    ) -> None:
        """Inject stimulator into given neuron groups.

        :meth:`Stimulator.connect_to_neuron_group` is called for each `group`.

        Parameters
        ----------
        stimulator : Stimulator
            The stimulator to inject
        *neuron_groups : NeuronGroup
            The groups to inject the stimulator into
        **kwparams : any
            Passed on to :meth:`Stimulator.connect_to_neuron_group` function.
            Necessary for parameters that can vary by neuron group, such
            as opsin expression levels.
        """
        self.inject_device(stimulator, *neuron_groups, **kwparams)
        self.stimulators[stimulator.name] = stimulator

    def inject_recorder(
        self, recorder: Recorder, *neuron_groups: NeuronGroup, **kwparams
    ) -> None:
        """Inject recorder into given neuron groups.

        :meth:`Recorder.connect_to_neuron_group` is called for each `group`.

        Parameters
        ----------
        recorder : Recorder
            The recorder to inject into the simulation
        *neuron_groups : NeuronGroup
            The groups to inject the recorder into
        **kwparams : any
            Passed on to :meth:`Recorder.connect_to_neuron_group` function.
            Necessary for parameters that can vary by neuron group, such
            as inhibitory/excitatory cell type
        """
        self.inject_device(recorder, *neuron_groups, **kwparams)
        self.recorders[recorder.name] = recorder

    def get_state(self) -> dict:
        """Return current recorder measurements.

        Returns
        -------
        dict
            A dictionary of `name`: `state` pairs for
            all recorders in the simulator.
        """
        state = {}
        for name, recorder in self.recorders.items():
            state[name] = recorder.get_state()
        return state

    def update_stimulators(self, ctrl_signals) -> None:
        """Update stimulators with output from the :class:`IOProcessor`

        Parameters
        ----------
        ctrl_signals : dict
            {`stimulator_name`: `ctrl_signal`} dictionary with values
            to update each stimulator.
        """
        if ctrl_signals is None:
            return
        for name, signal in ctrl_signals.items():
            self.stimulators[name].update(signal)

    def set_io_processor(self, io_processor, communication_period=None) -> None:
        """Set simulator IO processor

        Will replace any previous IOProcessor so there is only one at a time.
        A Brian NetworkOperation is created to govern communication between
        the Network and the IOProcessor.

        Parameters
        ----------
        io_processor : IOProcessor
        """
        self.io_processor = io_processor
        # remove previous NetworkOperation
        if self._processing_net_op is not None:
            self.network.remove(self._processing_net_op)
            self._processing_net_op = None

        if io_processor is None:
            return

        def communicate_with_io_proc(t):
            if io_processor.is_sampling_now(t / ms):
                io_processor.put_state(self.get_state(), t / ms)
            ctrl_signal = io_processor.get_ctrl_signal(t / ms)
            self.update_stimulators(ctrl_signal)

        # communication should be at every timestep. The IOProcessor
        # decides when to sample and deliver results.
        if communication_period is None:
            communication_period = defaultclock.dt
        self._processing_net_op = NetworkOperation(
            communicate_with_io_proc, dt=communication_period
        )
        self.network.add(self._processing_net_op)
        self.network.store(self._net_store_name)

    def run(self, duration: Quantity, **kwparams) -> None:
        """Run simulation.

        Parameters
        ----------
        duration : brian2 temporal Quantity
            Length of simulation
        **kwparams : additional arguments passed to brian2.run()
            level has a default value of 1
        """
        level = kwparams.get("level", 1)
        kwparams["level"] = level
        self.network.run(duration, **kwparams)

    def reset(self, **kwargs):
        """Reset the simulator to a neutral state

        Restores the Brian Network to where it was when the
        CLSimulator was last modified (last injection, IOProcessor change).
        Calls reset() on stimulators, recorders, and IOProcessor.
        """
        # kwargs passed to stimulators, recorders, and io_processor reset
        self.network.restore(self._net_store_name)
        for stim in self.stimulators.values():
            stim.reset(**kwargs)
        for rec in self.recorders.values():
            rec.reset(**kwargs)
        if self.io_processor is not None:
            self.io_processor.reset(**kwargs)



"""Assorted utilities for developers."""
from collections.abc import MutableMapping
from math import ceil, floor

from scipy import linalg
import numpy as np

from brian2 import second
from brian2.groups.group import get_dtype
from brian2.equations.equations import (
    Equations,
    DIFFERENTIAL_EQUATION,
    SUBEXPRESSION,
    PARAMETER,
)


def get_orth_vectors_for_v(v):
    """Returns w1, w2 as 1x3 row vectors"""
    q, r = linalg.qr(
        np.column_stack([v, v, v])
    )  # get two vectors orthogonal to v from QR decomp
    return q[:, 1], q[:, 2]


def xyz_from_rθz(rs, thetas, zs, xyz_start, xyz_end):
    """Convert from cylindrical to Cartesian coordinates."""
    # not using np.linalg.norm because it strips units
    cyl_length = np.sqrt(np.sum(np.subtract(xyz_end, xyz_start) ** 2))
    c = (xyz_end - xyz_start) / cyl_length  # unit vector in direction of cylinder

    r1, r2 = get_orth_vectors_for_v(c)

    def r_unit_vecs(thetas):
        cosines = np.reshape(np.cos(thetas), (len(thetas), 1))
        sines = np.reshape(np.sin(thetas), (len(thetas), 1))
        # add axis for broadcasting so result is nx3
        cosines = np.cos(thetas)[..., np.newaxis]
        sines = np.sin(thetas)[..., np.newaxis]
        return r1 * cosines + r2 * sines

    coords = (
        xyz_start + c * zs[..., np.newaxis] + rs[..., np.newaxis] * r_unit_vecs(thetas)
    )
    return coords[:, 0], coords[:, 1], coords[:, 2]


def uniform_cylinder_rθz(n, rmax, zmax):
    # generate Fibonacci spiral cylinder by rotating around axis
    # and up and down cylinder simultaneously, using different angles
    indices = np.arange(0, n) + 0.5
    rs = rmax * np.sqrt(indices / n)
    golden_angle = np.pi * (1 + np.sqrt(5))
    thetas = golden_angle * indices
    # using sqrt(2) instead of golden ratio here so
    # the two angles don't coincide
    phis = 2 * np.pi * np.sqrt(2) * indices
    zs = zmax * (1 + np.sin(phis)) / 2

    return rs, thetas, zs


def modify_model_with_eqs(neuron_group, eqs_to_add):
    """Adapted from _create_variables() from neurongroup.py from Brian2
    source code v2.3.0.2
    """
    if type(eqs_to_add) == str:
        eqs_to_add = Equations(eqs_to_add)
    neuron_group.equations += eqs_to_add

    variables = neuron_group.variables

    dtype = {}
    if isinstance(dtype, MutableMapping):
        dtype["lastspike"] = neuron_group._clock.variables["t"].dtype

    for eq in eqs_to_add.values():
        dtype = get_dtype(eq, dtype)  # {} appears to be the default
        if eq.type in (DIFFERENTIAL_EQUATION, PARAMETER):
            if "linked" in eq.flags:
                # 'linked' cannot be combined with other flags
                if not len(eq.flags) == 1:
                    raise SyntaxError(
                        ('The "linked" flag cannot be ' "combined with other flags")
                    )
                neuron_group._linked_variables.add(eq.varname)
            else:
                constant = "constant" in eq.flags
                shared = "shared" in eq.flags
                size = 1 if shared else neuron_group._N
                variables.add_array(
                    eq.varname,
                    size=size,
                    dimensions=eq.dim,
                    dtype=dtype,
                    constant=constant,
                    scalar=shared,
                )
        elif eq.type == SUBEXPRESSION:
            neuron_group.variables.add_subexpression(
                eq.varname,
                dimensions=eq.dim,
                expr=str(eq.expr),
                dtype=dtype,
                scalar="shared" in eq.flags,
            )
        else:
            raise AssertionError("Unknown type of equation: " + eq.eq_type)

    # Add the conditional-write attribute for variables with the
    # "unless refractory" flag
    if neuron_group._refractory is not False:
        for eq in neuron_group.equations.values():
            if eq.type == DIFFERENTIAL_EQUATION and "unless refractory" in eq.flags:
                not_refractory_var = neuron_group.variables["not_refractory"]
                var = neuron_group.variables[eq.varname]
                var.set_conditional_write(not_refractory_var)

    # Stochastic variables
    for xi in neuron_group.equations.stochastic_variables:
        neuron_group.variables.add_auxiliary_variable(
            xi, dimensions=(second**-0.5).dim
        )

    # Check scalar subexpressions
    for eq in neuron_group.equations.values():
        if eq.type == SUBEXPRESSION and "shared" in eq.flags:
            var = neuron_group.variables[eq.varname]
            for identifier in var.identifiers:
                if identifier in neuron_group.variables:
                    if not neuron_group.variables[identifier].scalar:
                        raise SyntaxError(
                            (
                                "Shared subexpression %s refers "
                                "to non-shared variable %s."
                            )
                            % (eq.varname, identifier)
                        )


def wavelength_to_rgb(wavelength_nm, gamma=0.8):
    """taken from http://www.noah.org/wiki/Wavelength_to_RGB_in_Python
    This converts a given wavelength of light to an
    approximate RGB color value. The wavelength must be given
    in nanometers in the range from 380 nm through 750 nm
    (789 THz through 400 THz).

    Based on code by Dan Bruton
    http://www.physics.sfasu.edu/astro/color/spectra.html
    """
    wavelength = wavelength_nm
    if wavelength < 380:
        wavelength = 380.0
    if wavelength > 750:
        wavelength = 750.0
    if wavelength >= 380 and wavelength <= 440:
        attenuation = 0.3 + 0.7 * (wavelength - 380) / (440 - 380)
        R = ((-(wavelength - 440) / (440 - 380)) * attenuation) ** gamma
        G = 0.0
        B = (1.0 * attenuation) ** gamma
    elif wavelength >= 440 and wavelength <= 490:
        R = 0.0
        G = ((wavelength - 440) / (490 - 440)) ** gamma
        B = 1.0
    elif wavelength >= 490 and wavelength <= 510:
        R = 0.0
        G = 1.0
        B = (-(wavelength - 510) / (510 - 490)) ** gamma
    elif wavelength >= 510 and wavelength <= 580:
        R = ((wavelength - 510) / (580 - 510)) ** gamma
        G = 1.0
        B = 0.0
    elif wavelength >= 580 and wavelength <= 645:
        R = 1.0
        G = (-(wavelength - 645) / (645 - 580)) ** gamma
        B = 0.0
    elif wavelength >= 645 and wavelength <= 750:
        attenuation = 0.3 + 0.7 * (750 - wavelength) / (750 - 645)
        R = (1.0 * attenuation) ** gamma
        G = 0.0
        B = 0.0
    else:
        R = 0.0
        G = 0.0
        B = 0.0
    return (R, G, B)



"""Contains a basic PI controller"""
from __future__ import annotations
from cleosim.processing import ProcessingBlock
from typing import Any


class PIController(ProcessingBlock):
    """Simple PI controller.

    :meth:`compute_output` requires a ``sample_time_ms`` keyword argument.
    Only tested on controlling scalar values, but could be easily adapted
    to controlling a multi-dimensional state.
    """

    ref_signal: callable[[float], Any]
    """Callable returning the target as a function of time in ms"""

    def __init__(
        self,
        ref_signal: callable,
        Kp: float,
        Ki: float = 0,
        sample_period_ms: float = 0,
        **kwargs: Any
    ):
        """
        Parameters
        ----------
        ref_signal : callable
            Must return the target as a function of time in ms
        Kp : float
            Gain on the proportional error
        Ki : float, optional
            Gain on the integral error, by default 0
        sample_period_ms : float, optional
            Rate at which processor takes samples, by default 0.
            Only used to compute integrated error on first sample
        """
        super().__init__(**kwargs)
        self.ref_signal = ref_signal
        self.Kp = Kp
        self.Ki = Ki
        self.sample_period_ms = sample_period_ms
        self.integrated_error = 0
        self.prev_time_ms = None

    def compute_output(self, input: float, **kwargs) -> float:
        """Compute control input to the system using previously specified gains.

        Parameters
        ----------
        input : Any
            Current system state

        Returns
        -------
        float
            Control signal
        """
        time_ms = kwargs["sample_time_ms"]
        if self.prev_time_ms is None:
            self.prev_time_ms = time_ms - self.sample_period_ms
        intersample_period_s = (time_ms - self.prev_time_ms) / 1000
        error = self.ref_signal(time_ms) - input
        self.integrated_error += error * intersample_period_s
        self.prev_time_ms = time_ms
        return self.Kp * error + self.Ki * self.integrated_error



"""Contains probes, coordinate convenience functions, signals, spiking, and LFP"""
from cleosim.electrodes.lfp import TKLFPSignal
from cleosim.electrodes.spiking import MultiUnitSpiking, SortedSpiking, Spiking
from cleosim.electrodes.probes import (
    Probe,
    Signal,
    concat_coords,
    linear_shank_coords,
    tetrode_shank_coords,
    poly2_shank_coords,
    poly3_shank_coords,
    tile_coords,
)



"""Classes and functions for constructing and configuring an :class:`~cleosim.IOProcessor`."""
from cleosim.processing.base import (
    LatencyIOProcessor,
    ProcessingBlock,
    RecordOnlyProcessor,
)
from cleosim.processing.controllers import PIController
from cleosim.processing.observers import FiringRateEstimator
from cleosim.processing.delays import Delay, ConstantDelay, GaussianDelay
